import asyncio
import functools
import logging
import socket
import time
from concurrent.futures.thread import ThreadPoolExecutor
from enum import Enum

import requests
from django.conf import settings
from requests import RequestException


class WebhookStatus(Enum):
    FAILED = "failed"
    SUCCEEDED = "succeeded"
    WARNING = "warning"


class Webhook:
    """
    Base class for webhooks in the registry. Run function initiates sending of webhook to the specified url.
    """

    code = None
    data = {}

    def __init__(self, code=None):
        if code:
            self.code = code

    @classmethod
    def get_from_registry(cls):

        """
        Getting webhook from registry using the code.
        """

        from hurricane.webhooks import webhook_registry

        return webhook_registry.get(cls.code)

    def run(self, url: str, status: WebhookStatus, error_trace: str = None, close_loop: bool = False, loop=None):

        """
        Initiates the sending of webhook in an asynchronous manner. Also specifies the callback of the async process,
        which handles the feedback and either logs success or failure of a webhook sending process.

        Parameters
        ----------
        url : Url, which webhook should be sent to
        status : can be either WebhookStatus.FAILED or WebhookStatus.SUCCEEDED depending on the success or failure of
        the process, which should be indicated by the webhook
        error_trace : specifies the error trace of the preceding failure
        close_loop : specifies, whether the main loop should be closed or be left running
        """

        if url:
            self.set_traceback(error_trace)
            self.set_status(status)
            self.set_timestamp()
            self.set_podname()
            self.set_version()
            current_loop = loop or asyncio.get_event_loop()
            executor = ThreadPoolExecutor(max_workers=1)
            fut = current_loop.run_in_executor(executor, self._send_webhook, self.get_message(), url, close_loop)
            # callback runs after run_in_executor is done
            callback_wrapper = functools.partial(
                self._callback_webhook_exception_check, url=url, close_loop=close_loop, loop=loop
            )
            fut.add_done_callback(callback_wrapper)

    def _send_webhook(self, data: dict, webhook_url: str, close_loop: bool):
        # sending webhook request to the specified url
        logger = logging.getLogger()
        logger.info(f"Start sending {self.code} webhook to {webhook_url}")

        response = requests.post(webhook_url, timeout=5, json=data)
        response.raise_for_status()

    def set_traceback(self, error_trace: str):
        self.data["traceback"] = error_trace

    def set_timestamp(self):
        self.data["timestamp"] = int(time.time())

    def set_podname(self):
        self.data["podname"] = socket.gethostname()

    def set_status(self, status: WebhookStatus):
        self.data["status"] = status.value

    def set_version(self):
        if hasattr(settings, "HURRICANE_VERSION"):
            self.data["version"] = settings.HURRICANE_VERSION
        else:
            self.data["version"] = None

    def get_message(self):
        return self.data

    @staticmethod
    def _callback_webhook_exception_check(future: asyncio.Future, url: str, close_loop: bool, loop=None):
        # checks if sending webhook had any failures, it indicates, that command was successfully executed
        # but sending webhook has failed
        logger = logging.getLogger()
        try:
            future.result()
        except RequestException as e:
            logger.warning(f"Sending webhook to {url} has failed due to {e}")

        if close_loop:
            logger.info("Loop will be closed")
            loop.stop()



import os
import socket
import subprocess
from queue import Empty, Queue
from threading import Thread
from time import sleep
from typing import List, Tuple, Union

import docker

from hurricane.testing.actors import TestPublisher

MANAGE_FILE = "manage.py"


class BusyPortException(Exception):
    pass


class HurricaneBaseDriver(object):
    proc = None
    log_lines = []
    base_command = []
    coverage_base_command = []
    test_string = ""
    ports = [8000, 8001]
    source = "--source=hurricane/"

    def __init__(self):
        for port in self.ports:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                try:
                    sock.bind(("127.0.0.1", port))
                except OSError:
                    raise BusyPortException(f"Port {port} already in use.")

    def get_server_host_port(self, probe_port=False) -> Union[Tuple[str, int], Tuple[None, None]]:
        port = self.probe_port if probe_port else self.port
        if self.proc:
            return "localhost", port
        else:
            return None, None

    def get_output(self, read_all=False) -> Tuple[str, str]:
        if read_all:
            while True:
                try:
                    line = self.q.get(timeout=0.5)
                    self.log_lines.append(line)
                except Empty:
                    break
            return "".join(self.log_lines), ""
        else:
            try:
                line = self.q.get(timeout=1)
                if line:
                    return line, ""
            except Empty:
                pass

    def _get_env(self):
        return os.environ.copy()

    def _start(self, params: List[str] = None, coverage: bool = True) -> None:
        self.log_lines = []
        base_command = self.coverage_base_command if coverage else self.base_command

        def enqueue_stdout(proc, queue):
            out = proc.stdout
            for line in iter(out.readline, b""):
                queue.put(line.decode("utf-8"))
            out.close()

        def enqueue_stderr(proc, queue):
            out = proc.stderr
            for line in iter(out.readline, b""):
                queue.put(line.decode("utf-8"))
            out.close()

        params = params if params else []

        base_command = base_command + params
        if "--port" in params:
            self.port = int(params[params.index("--port") + 1])
        else:
            self.port = 8000

        if "--probe-port" in params:
            self.probe_port = int(params[params.index("--probe-port") + 1])
        else:
            self.probe_port = 8001

        self.proc = subprocess.Popen(base_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=self._get_env())
        self.q = Queue()
        self.t_stderr = Thread(target=enqueue_stderr, args=(self.proc, self.q))
        self.t_stdout = Thread(target=enqueue_stdout, args=(self.proc, self.q))
        self.t_stderr.daemon = True
        self.t_stderr.start()
        self.t_stdout.daemon = True
        self.t_stdout.start()
        # wait a maximum of 1 second
        for _ in range(10):
            if self.proc:
                out, err = self.get_output(read_all=True)
            else:
                out = ""
            if self.test_string in out:
                break
            sleep(0.1)

    def _stop(self):
        if self.proc:
            self.proc.terminate()


class HurricaneServerDriver(HurricaneBaseDriver):
    coverage_base_command = [
        "coverage",
        "run",
        HurricaneBaseDriver.source,
        MANAGE_FILE,
        "serve",
    ]
    base_command = ["python", MANAGE_FILE, "serve"]
    test_string = "Tornado-powered Django web server"
    _env = {}

    def _get_env(self):
        env = super(HurricaneServerDriver, self)._get_env()
        env.update(self._env)
        return env

    def start_server(self, params: dict = None, coverage: bool = True, env: dict = None) -> None:
        self._env = env or dict()
        self._start(params, coverage)

    def stop_server(self) -> None:
        self._stop()


class HurricaneWebhookServerDriver(HurricaneBaseDriver):
    ports = [8040, 8041]
    coverage_base_command = [
        "coverage",
        "run",
        "-m",
        HurricaneBaseDriver.source,
        "hurricane.testing.start_webhook_receiver",
    ]
    base_command = ["python", "-m", "hurricane.testing.start_webhook_receiver"]
    test_string = "Started webhook receiver server"
    _env = {}

    def _get_env(self):
        env = super(HurricaneWebhookServerDriver, self)._get_env()
        env.update(self._env)
        return env

    def start_server(self, params: dict = None, coverage: bool = True, env: dict = None) -> None:
        self._env = env or dict()
        self._start(params, coverage)

    def stop_server(self) -> None:
        self._stop()


class HurricaneK8sServerDriver(HurricaneBaseDriver):
    ports = [8040, 8041]
    coverage_base_command = [
        "coverage",
        "run",
        "-m",
        HurricaneBaseDriver.source,
        "hurricane.testing.start_k8s_server",
    ]
    base_command = ["python", "-m", "hurricane.testing.start_k8s_server"]
    test_string = "Started K8s server"
    _env = {}

    def _get_env(self):
        env = super(HurricaneK8sServerDriver, self)._get_env()
        env.update(self._env)
        return env

    def start_server(self, params: dict = None, coverage: bool = True, env: dict = None) -> None:
        self._env = env
        self._start(params, coverage)

    def stop_server(self) -> None:
        self._stop()


class HurricaneAMQPDriver(HurricaneBaseDriver):
    coverage_base_command = [
        "coverage",
        "run",
        HurricaneBaseDriver.source,
        MANAGE_FILE,
        "consume",
    ]
    base_command = ["python", MANAGE_FILE, "consume"]
    test_string = "Starting a Tornado-powered Django AMQP consumer"
    ports = [5672, 8000, 8001]
    _env = {}

    def _get_env(self):
        env = super(HurricaneAMQPDriver, self)._get_env()
        env.update(self._env)
        return env

    def start_amqp(self) -> None:
        client = docker.from_env()
        if hasattr(self, "container") and self.container:
            c = client.containers.run(
                "quay.io/blueshoe/rabbitmq3.8-alpine",
                auto_remove=True,
                detach=True,
                ports={"5672": ("127.0.0.1", self._temp_port)},
            )
        else:
            c = client.containers.run(
                "quay.io/blueshoe/rabbitmq3.8-alpine",
                auto_remove=True,
                detach=True,
                ports={"5672": ("127.0.0.1", None)},
            )
        self.container = client.containers.get(c.id)
        # busy wait for rabbitmq to come up (timeout 10 seconds)
        for _ in range(20):
            if "Ready to start client connection listeners" in self.container.logs().decode("utf-8"):
                break
            else:
                sleep(0.5)

    def get_test_publisher(self, vhost="/"):
        host, port = self.get_amqp_host_port()
        return TestPublisher(host, port, vhost)

    def start_consumer(self, params: List[str] = None, coverage: bool = True, env: dict = None) -> None:
        self._env = env
        self._start(params, coverage)

    def stop_amqp(self) -> None:
        if hasattr(self, "container") and self.container:
            try:
                self.container.kill()
                delattr(self, "container")
            except Exception:
                # this container is potentially already stopped
                delattr(self, "container")

    def halt_amqp(self) -> None:
        if hasattr(self, "container") and self.container:
            try:
                _, self._temp_port = self.get_amqp_host_port()
                self.container.kill()
            except Exception:
                # this container is potentially already stopped
                pass

    def stop_consumer(self) -> None:
        self._stop()

    def _get_port(self):
        if hasattr(self, "container") and self.container:
            self._temp_port = self.container.attrs["NetworkSettings"]["Ports"]["5672/tcp"][0]["HostPort"]
            return self._temp_port
        return None

    def get_amqp_host_port(self) -> Union[Tuple[str, int], Tuple[None, None]]:
        if port := self._get_port():
            return "127.0.0.1", port
        else:
            return None, None



import asyncio
import signal
import sys
import time
import traceback
from typing import Callable

import tornado
from django.conf import settings
from django.core.management import call_command
from django.db import connections
from django.db.migrations.executor import MigrationExecutor

from hurricane.metrics import RequestCounterMetric, ResponseTimeAverageMetric, StartupTimeMetric
from hurricane.server.django import DjangoHandler, DjangoLivenessHandler, DjangoReadinessHandler, DjangoStartupHandler
from hurricane.server.loggers import access_log, logger
from hurricane.webhooks import StartupWebhook
from hurricane.webhooks.base import WebhookStatus


class HurricaneApplication(tornado.web.Application):
    def __init__(self, *args, **kwargs):
        self.collect_metrics = True
        if "metrics" in kwargs:
            self.collect_metrics = kwargs["metrics"]
        super(HurricaneApplication, self).__init__(*args, **kwargs)

    def log_request(self, handler: DjangoHandler) -> None:
        """Writes a completed HTTP request to the logs."""
        if handler.get_status() < 400:
            log_method = access_log.info
        elif handler.get_status() < 500:
            log_method = access_log.warning
        else:
            log_method = access_log.error
        request_time = 1000.0 * handler.request.request_time()
        log_method(
            "%d %s %.2fms",
            handler.get_status(),
            handler._request_summary(),
            request_time,
        )
        if self.collect_metrics:
            RequestCounterMetric.increment()
            ResponseTimeAverageMetric.add_value(request_time)


def make_probe_server(options, check_func):
    """create probe route application"""
    handlers = [
        (
            options["liveness_probe"],
            DjangoLivenessHandler,
            {"check_handler": check_func, "webhook_url": options["webhook_url"]},
        ),
        (
            options["readiness_probe"],
            DjangoReadinessHandler,
            {"req_queue_len": options["req_queue_len"], "webhook_url": options["webhook_url"]},
        ),
        (options["startup_probe"], DjangoStartupHandler),
    ]
    return HurricaneApplication(handlers, debug=options["debug"], metrics=False)


def make_http_server(options, check_func, include_probe=False):
    """create all routes for this application"""
    if include_probe:
        handlers = [
            (options["liveness_probe"], DjangoLivenessHandler, {"check_handler": check_func}),
            (options["readiness_probe"], DjangoReadinessHandler, {"req_queue_len": options["req_queue_len"]}),
            (options["startup_probe"], DjangoStartupHandler),
        ]
    else:
        handlers = []
    # if static file serving is enabled
    if options["static"]:
        logger.info(f"Serving static files under {settings.STATIC_URL} from {settings.STATIC_ROOT}")
        handlers.append(
            (
                f"{settings.STATIC_URL}(.*)",
                tornado.web.StaticFileHandler,
                {"path": settings.STATIC_ROOT},
            )
        )
    # if media file serving is enabled
    if options["media"]:
        logger.info(f"Serving media files under {settings.MEDIA_URL} from {settings.MEDIA_ROOT}")
        handlers.append(
            (
                f"{settings.MEDIA_URL}(.*)",
                tornado.web.StaticFileHandler,
                {"path": settings.MEDIA_ROOT},
            )
        )

    # append the django routing system
    handlers.append((".*", DjangoHandler))
    return HurricaneApplication(handlers, debug=options["debug"], metrics=not options["no_metrics"])


def make_http_server_and_listen(start_time: float, options: dict, check: Callable, include_probe: bool) -> None:
    logger.info(f"Starting HTTP Server on port {options['port']}")
    django_application = make_http_server(options, check, include_probe)
    django_application.listen(options["port"])
    StartupWebhook().run(url=options["webhook_url"] or None, status=WebhookStatus.SUCCEEDED)
    end_time = time.time()
    time_elapsed = end_time - start_time
    # if startup time metric value is set - startup process is finished
    StartupTimeMetric.set(time_elapsed)
    logger.info(f"Startup time is {time_elapsed} seconds")


def command_task(commands: list, webhook_url: str = None, loop: asyncio.unix_events.SelectorEventLoop = None) -> None:
    logger.info("Starting execution of management commands")
    for command in commands:

        # split a command string to get command options
        command_split = command[0].split()
        logger.info(f"Starting execution of command {command_split[0]} with arguments {command_split[1:]}")
        start_time_command = time.time()
        # call management command
        try:
            call_command(*command_split)
        except Exception as e:
            logger.error(e)
            error_trace = traceback.format_exc()
            logger.info("Webhook with a status failed has been initiated")
            # webhook is registered and run in a new thread, not blocking the process
            StartupWebhook().run(
                url=webhook_url or None,
                error_trace=error_trace,
                close_loop=True,
                status=WebhookStatus.FAILED,
                loop=loop,
            )
            raise e

        end_time_command = time.time()

        logger.info(f"Command {command_split[0]} was executed in {end_time_command - start_time_command} seconds")


def check_databases():
    for db_name in connections:
        connection = connections[db_name]
        cursor = connection.cursor()
        try:
            cursor.execute("SELECT (1)")
            logger.info("Database was checked successfully")
            cursor.close()
            return True
        except Exception as e:
            logger.warning(f"Database command execution has failed with {e}")
            cursor.close()
            return False


def count_migrations():
    number_of_migrations = 0
    for db_name in connections:
        connection = connections[db_name]
        if hasattr(connection, "prepare_database"):
            connection.prepare_database()
        executor = MigrationExecutor(connection)
        targets = executor.loader.graph.leaf_nodes()
        number_of_migrations += len(executor.migration_plan(targets))
    return number_of_migrations


def signal_handler(signal, frame):
    logger.error("\nprogram exiting gracefully")
    sys.exit(0)


signal.signal(signal.SIGINT, signal_handler)


def check_db_and_migrations(webhook_url: str = None, loop: asyncio.unix_events.SelectorEventLoop = None):
    try:
        while check_databases():
            number_of_migrations = count_migrations()

            logger.info(f"There are {number_of_migrations} pending migrations")
            if number_of_migrations == 0:
                logger.info("No pending migrations")
                break

    except Exception as e:
        logger.error(e)
        error_trace = traceback.format_exc()
        logger.info("Webhook with a status warning has been initiated")

        StartupWebhook().run(
            url=webhook_url or None, error_trace=error_trace, close_loop=False, status=WebhookStatus.WARNING, loop=loop
        )



import logging

import tornado.ioloop

from hurricane.testing.actors import K8sServer


def start_server():
    logging.info("Started K8s server")
    app = K8sServer().make_http_receiver_app()
    app.listen(8072)
    tornado.ioloop.IOLoop.current().start()


if __name__ == "__main__":
    start_server()



import asyncio
import functools
import signal
import time
import traceback
from concurrent.futures.thread import ThreadPoolExecutor

import tornado.autoreload
import tornado.web
import tornado.wsgi
from django.core.management.base import BaseCommand

from hurricane.server import (
    check_db_and_migrations,
    command_task,
    logger,
    make_http_server_and_listen,
    make_probe_server,
)
from hurricane.webhooks import StartupWebhook
from hurricane.webhooks.base import WebhookStatus


class Command(BaseCommand):

    """
    Start a Tornado-powered Django web server.
    Implements serve command as a management command for django application.
    The new command can be called using ``python manage.py server <arguments>``.
    It also can take command arguments, which are python django management commands and will be executed asynchronously.
    Upon successful execution of management commands, application server will be started. During execution of management
    commands probe server can be polled, in particular startup probe, which will respond with a status 400.
    Arguments:
        - ``--static`` - serve collected static files
        - ``--media`` - serve media files
        - ``--autoreload`` - reload code on change
        - ``--debug`` - set Tornado's Debug flag
        - ``--port`` - the port for Tornado to listen on
        - ``--startup-probe`` - the exposed path (default is /startup) for probes to check startup
        - ``--readiness-probe`` - the exposed path (default is /ready) for probes to check readiness
        - ``--liveness-probe`` - the exposed path (default is /alive) for probes to check liveness
        - ``--probe-port`` - the port for Tornado probe route to listen on
        - ``--req-queue-len`` - threshold of length of queue of request, which is considered for readiness probe
        - ``--no-probe`` - disable probe endpoint
        - ``--no-metrics`` - disable metrics collection
        - ``--command`` - repetitive command for adding execution of management commands before serving
        - ``--check-migrations`` - check if all migrations were applied before starting application
        - ``--webhook-url``- If specified, webhooks will be sent to this url
    """

    help = "Start a Tornado-powered Django web server"

    def add_arguments(self, parser):
        """
        Defines arguments, that can be accepted with ``serve`` command.
        """
        parser.add_argument("--static", action="store_true", help="Serve collected static files")
        parser.add_argument("--media", action="store_true", help="Serve media files")
        parser.add_argument("--autoreload", action="store_true", help="Reload code on change")
        parser.add_argument("--debug", action="store_true", help="Set Tornado's Debug flag")
        parser.add_argument("--port", type=int, default=8000, help="The port for Tornado to listen on")
        parser.add_argument(
            "--liveness-probe",
            type=str,
            default="/alive",
            help="The exposed path (default is /alive) for probes to check liveness",
        )
        parser.add_argument(
            "--readiness-probe",
            type=str,
            default="/ready",
            help="The exposed path (default is /ready) for probes to check readiness",
        )
        parser.add_argument(
            "--startup-probe",
            type=str,
            default="/startup",
            help="The exposed path (default is /startup) for probes to check startup",
        )
        parser.add_argument(
            "--probe-port",
            type=int,
            help="The port for Tornado probe route to listen on",
        )
        parser.add_argument("--req-queue-len", type=int, default=10, help="Length of the request queue")
        parser.add_argument("--no-probe", action="store_true", help="Disable probe endpoint")
        parser.add_argument("--no-metrics", action="store_true", help="Disable metrics collection")
        parser.add_argument("--command", type=str, action="append", nargs="+")
        parser.add_argument("--check-migrations", action="store_true", help="Check if migrations were applied")
        parser.add_argument(
            "--webhook-url",
            type=str,
            help="Url for webhooks",
        )

    def handle(self, *args, **options):
        """
        Defines functionalities for different arguments. After all arguments were processed, it starts the async event
        loop.
        """
        start_time = time.time()
        logger.info("Tornado-powered Django web server")

        if options["autoreload"]:
            tornado.autoreload.start()
            logger.info("Autoreload was performed")

        # set the probe port
        # the probe port by default is supposed to run the next port of the application
        probe_port = options["probe_port"] if options["probe_port"] else options["port"] + 1

        # sanitize probe paths
        options["liveness_probe"] = f"/{options['liveness_probe'].lstrip('/')}"
        options["readiness_probe"] = f"/{options['readiness_probe'].lstrip('/')}"
        options["startup_probe"] = f"/{options['startup_probe'].lstrip('/')}"

        # set the probe routes
        # if probe port is set to the application's port, add it to the application's routes
        include_probe = False
        if not options["no_probe"]:
            if probe_port != options["port"]:
                logger.info(
                    f"Starting probe application running on port {probe_port} with route liveness-probe: "
                    f"{options['liveness_probe']}, readiness-probe: {options['readiness_probe']}, "
                    f"startup-probe: {options['startup_probe']}"
                )
                probe_application = make_probe_server(options, self.check)
                probe_application.listen(probe_port)
            else:
                include_probe = True
                logger.info(
                    f"Starting probe application with routes liveness-probe: {options['liveness_probe']}, "
                    f"readiness-probe: {options['readiness_probe']}, startup-probe: {options['startup_probe']} "
                    f"running integrated on port {probe_port}"
                )

        else:
            logger.info("No probe application running")

        loop = asyncio.get_event_loop()

        make_http_server_wrapper = functools.partial(
            make_http_server_and_listen,
            start_time=start_time,
            options=options,
            check=self.check,
            include_probe=include_probe,
        )

        # all commands, that should be executed before starting http server should be added to this list
        exec_list = []
        if options["command"]:
            # wrap function to use additional variables
            management_commands_wrapper = functools.partial(
                command_task,
                commands=options["command"],
                webhook_url=options["webhook_url"] or None,
                loop=loop,
            )
            exec_list.append(management_commands_wrapper)
        if options["check_migrations"]:
            check_db_and_migrations_wrapper = functools.partial(
                check_db_and_migrations, webhook_url=options["webhook_url"] or None, loop=loop
            )
            exec_list.append(check_db_and_migrations_wrapper)

        def bundle_func(exec_list, loop, make_http_server_wrapper):
            # executes functions from exec_list sequentially
            for func in exec_list:
                func(*args)
            # after all functions were executed http server is started in the main thread
            # call_soon_threadsafe puts http server into the main loop of the program
            loop.call_soon_threadsafe(make_http_server_wrapper)

        executor = ThreadPoolExecutor(max_workers=1)
        # bundle_func is executed in a separate thread. Main thread has an active probe server, which should not be
        # interrupted
        loop.run_in_executor(executor, bundle_func, exec_list, loop, make_http_server_wrapper)

        def ask_exit(signame):
            logger.info(f"Received signal {signame}. Shutting down now.")
            loop.stop()

        for signame in ("SIGINT", "SIGTERM"):
            loop.add_signal_handler(getattr(signal, signame), functools.partial(ask_exit, signame))

        loop.run_forever()



import asyncio
import logging

import tornado.web


class K8sServerMetricsHandler(tornado.web.RequestHandler):
    """This handler reports current usage statistics to Kubernetes"""

    logger = logging.getLogger()

    def get(self):
        # write custom metrics to MetricsAPI in the future
        request_queue_length = len(asyncio.all_tasks())
        self.logger.info(f"Request count: {request_queue_length}")
        self.write(f"{request_queue_length}".encode())
